
// includes
#include "../gfx/opengl.hpp"
#include "../img/Image.hpp"
#include "../platform/Window.hpp"
#include "../debug.hpp"
#include "Texture.hpp"

using namespace bleifrei::img;
using namespace bleifrei::platform;

namespace bleifrei {
	namespace tex {

		unsigned int Texture::m_currentType = 0;
		int Texture::m_currentUnit = 0;

		Texture *Texture::m_textureUnit[MaxTextureUnits] = { 0, 0, 0, 0, 0, 0, 0, 0 };

		Texture::Texture(unsigned int type) :
			m_type(type)
		{
		}

		void
		Texture::create(void)
		{
			glGenTextures(1, &m_id);
		}

		void
		Texture::destroy(void)
		{
			if (this == m_textureUnit[m_currentUnit])
			{
				nullTexture->bind(m_currentUnit);
			}
			glDeleteTextures(1, &m_id);
		}

		void
		Texture::bind(unsigned int unit)
		{
			setUnit(unit);
			Texture *texture = m_textureUnit[m_currentUnit];
			if ( texture != this )
			{
				if ( texture != 0 && ( this == 0 || texture->m_type != m_type ) )
				{
					texture->disable();
				}
				if ( this != 0 ) {
					if (!texture || texture->m_type != m_type) enable();
					glBindTexture(m_type, m_id);
				}
				m_textureUnit[m_currentUnit] = this;
			}
		}

		void
		Texture::forceBind(unsigned int unit)
		{
			setUnit(unit);
			glDisable(GL_TEXTURE_1D);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_TEXTURE_3D);
			glDisable(GL_TEXTURE_CUBE_MAP);
			glDisable(GL_TEXTURE_RECTANGLE_NV);

			if (this)
			{
				glEnable(m_type);
				glBindTexture(m_type, m_id);
			}
			m_textureUnit[unit] = this;
		}

		void
		Texture::reset(void)
		{
			for (int i = 0; i < MaxTextureUnits; ++i)
			{
				setUnit(i);
				glDisable(GL_TEXTURE_1D);
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_TEXTURE_3D);
				glDisable(GL_TEXTURE_CUBE_MAP);
				glDisable(GL_TEXTURE_RECTANGLE_NV);
				m_textureUnit[i] = nullTexture;
			}
			setUnit(0);
		}

		void
		Texture::loadFromFramebuffer(int x, int y, int width, int height)
		{
			bind();
			glCopyTexImage2D(m_type, 0, GL_RGBA, x, y, width, height, 0);
		}

		void
		Texture::setUnit(unsigned int unit)
		{
			if (m_currentUnit != unit)
			{
				BLEIFREI_ASSERT( unit >= 0 && unit < MaxTextureUnits );
				glActiveTexture(GL_TEXTURE0 + unit);
				m_currentUnit = unit;
			}
		}

		void
		Texture::setMinFilter(TextureFilter filter)
		{
			bind();

			BLEIFREI_ASSERT(this != 0);
			BLEIFREI_ASSERT( filter < TextureFilterMax )
			if (filter == TextureFilterBilinear) {
				glTexParameteri(m_type, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			}
			else if (filter == TextureFilterTrilinear) {
				glTexParameteri(m_type, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			}
			else {
				glTexParameteri(m_type, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			}
		}

		void
		Texture::setMagFilter(TextureFilter filter)
		{
			bind();

			BLEIFREI_ASSERT(this != 0)
			BLEIFREI_ASSERT( filter != TextureFilterTrilinear )
			BLEIFREI_ASSERT( filter < TextureFilterMax )
			if (filter == TextureFilterBilinear) {
				glTexParameteri(m_type, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			}
			else {
				glTexParameteri(m_type, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			}
		}

		void
		Texture::setAnisotropicFiltering(float level)
		{
			bind();

			BLEIFREI_ASSERT(this != 0);
			float maxlevel;
			glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxlevel);

			float anisotropy = (level <= maxlevel) ? level : maxlevel;
			glTexParameterfv(m_type, GL_TEXTURE_MAX_ANISOTROPY_EXT, &anisotropy);
		}

		void
		Texture::enable(void)
		{
			BLEIFREI_ASSERT( this != 0 );
			glEnable(m_type);
			m_currentType = m_type;
		}

		void
		Texture::disable(void)
		{
			BLEIFREI_ASSERT(this != 0);
			glDisable(m_type);
		}

		Texture2::Texture2(void) :
			Texture(GL_TEXTURE_2D)
		{
			create();
		}

		Texture2::~Texture2(void)
		{
			destroy();
		}

		void
		Texture2::load(Image *image)
		{
			Texture *oldTexture = m_textureUnit[m_currentUnit];
			bind();

			// $TODO: change to new texture friendly image format
			// $TODO: handle enabled/disabled mipmap
			switch (image->getFormat()) {
			case ImageFormatL8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 1, image->getWidth(), image->getHeight(), GL_LUMINANCE, GL_UNSIGNED_BYTE, image->getData());
				break;
			case ImageFormatA8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 1, image->getWidth(), image->getHeight(), GL_ALPHA, GL_UNSIGNED_BYTE, image->getData());
				break;
			//case 16 :
				//glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE16, image.getWidth(), image.getHeight(), 0, GL_LUMINANCE, GL_UNSIGNED_SHORT, image->getData());
				//gluBuild2DMipmaps(GL_TEXTURE_2D, 1, image.getWidth(), image.getHeight(), GL_LUMINANCE16, GL_UNSIGNED_SHORT, image.getData());
				//break;
			case ImageFormatR8G8B8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 3, image->getWidth(), image->getHeight(), GL_RGB, GL_UNSIGNED_BYTE, image->getData());
				break;
			case ImageFormatB8G8R8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 3, image->getWidth(), image->getHeight(), GL_BGR, GL_UNSIGNED_BYTE, image->getData());
				break;
			case ImageFormatR8G8B8A8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 4, image->getWidth(), image->getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, image->getData());
				break;
			case ImageFormatB8G8R8A8 :
				gluBuild2DMipmaps(GL_TEXTURE_2D, 4, image->getWidth(), image->getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, image->getData());
				break;
			default:
				BLEIFREI_ASSERT(false);
			}

			oldTexture->bind();
		}

		TextureRect::TextureRect(void) :
			Texture(GL_TEXTURE_RECTANGLE_NV)
		{
			create();
		}

		TextureRect::~TextureRect(void)
		{
			destroy();
		}

		TextureCube::TextureCube(void) :
			Texture(GL_TEXTURE_CUBE_MAP)
		{
			create();
		}

		TextureCube::~TextureCube(void)
		{
			destroy();
		}

		void
		TextureCube::load(Image &neg_x, Image &pos_x, Image &neg_y, Image &pos_y, Image &neg_z, Image &pos_z)
		{
			Texture *oldTexture = m_textureUnit[m_currentUnit];
			bind();

			glTexParameteri(m_type, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(m_type, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			glTexParameteri(m_type, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

			if (neg_x.getFormat() != pos_x.getFormat() ||
				pos_x.getFormat() != neg_y.getFormat() ||
				neg_y.getFormat() != pos_y.getFormat() ||
				pos_y.getFormat() != neg_z.getFormat() ||
				neg_z.getFormat() != pos_z.getFormat())
			{
				// $TODO: throw exception. must have same format
				return;
			}

			switch (neg_x.getFormat())
			{
			case ImageFormatL8 :
				glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, GL_LUMINANCE, neg_x.getWidth(), neg_x.getHeight(), 0, 1, GL_UNSIGNED_BYTE, neg_x.getData());
				glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_LUMINANCE, pos_x.getWidth(), pos_x.getHeight(), 0, 1, GL_UNSIGNED_BYTE, pos_x.getData());
				glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, GL_LUMINANCE, neg_y.getWidth(), neg_y.getHeight(), 0, 1, GL_UNSIGNED_BYTE, neg_y.getData());
				glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, GL_LUMINANCE, pos_y.getWidth(), pos_y.getHeight(), 0, 1, GL_UNSIGNED_BYTE, pos_y.getData());
				glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, GL_LUMINANCE, neg_z.getWidth(), neg_z.getHeight(), 0, 1, GL_UNSIGNED_BYTE, neg_z.getData());
				glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, GL_LUMINANCE, pos_z.getWidth(), pos_z.getHeight(), 0, 1, GL_UNSIGNED_BYTE, pos_z.getData());
				break;
			case ImageFormatR8G8B8 :
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 3, neg_x.getWidth(), neg_x.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, neg_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 3, pos_x.getWidth(), pos_x.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, pos_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 3, neg_y.getWidth(), neg_y.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, neg_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 3, pos_y.getWidth(), pos_y.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, pos_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 3, neg_z.getWidth(), neg_z.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, neg_z.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 3, pos_z.getWidth(), pos_z.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, pos_z.getData());
				break;
			case ImageFormatB8G8R8 :
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 3, neg_x.getWidth(), neg_x.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, neg_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 3, pos_x.getWidth(), pos_x.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, pos_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 3, neg_y.getWidth(), neg_y.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, neg_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 3, pos_y.getWidth(), pos_y.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, pos_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 3, neg_z.getWidth(), neg_z.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, neg_z.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 3, pos_z.getWidth(), pos_z.getHeight(), GL_BGR, GL_UNSIGNED_BYTE, pos_z.getData());
				break;
			case ImageFormatR8G8B8A8 :
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 4, neg_x.getWidth(), neg_x.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, neg_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 4, pos_x.getWidth(), pos_x.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, pos_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 4, neg_y.getWidth(), neg_y.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, neg_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 4, pos_y.getWidth(), pos_y.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, pos_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 4, neg_z.getWidth(), neg_z.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, neg_z.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 4, pos_z.getWidth(), pos_z.getHeight(), GL_RGBA, GL_UNSIGNED_BYTE, pos_z.getData());
				break;
			case ImageFormatB8G8R8A8 :
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 4, neg_x.getWidth(), neg_x.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, neg_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 4, pos_x.getWidth(), pos_x.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, pos_x.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 4, neg_y.getWidth(), neg_y.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, neg_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 4, pos_y.getWidth(), pos_y.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, pos_y.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 4, neg_z.getWidth(), neg_z.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, neg_z.getData());
				gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 4, pos_z.getWidth(), pos_z.getHeight(), GL_BGRA, GL_UNSIGNED_BYTE, pos_z.getData());
				break;
			}
		}

	}
}